% Copyright (c) Camden Dixie O'Brien
% SPDX-License-Identifier: AGPL-3.0-only

-module(tcp_server).
-behaviour(gen_server).

-export([start_link/2]).
-export([init/1, handle_call/3, handle_cast/2, handle_info/2,
		 terminate/2, code_change/3]).

-record(state, {socket, acceptor}).

start_link(Port, CertDir) ->
	gen_server:start_link({local, ?MODULE}, ?MODULE, [Port, CertDir], []).

init([Port, CertDir]) ->
	TcpOpts = [binary, inet6, {active, false}, {reuseaddr, true}],
	SslOpts = [{certfile, filename:join([CertDir, "server.pem"])},
			   {cacertfile, filename:join([CertDir, "ca.pem"])},
			   {verify, verify_peer},
			   {fail_if_no_peer_cert, true}],
	{ok, Socket} = ssl:listen(Port, TcpOpts ++ SslOpts),
	Pid = spawn_link(fun() -> acceptor_loop(Socket) end),
	{ok, #state{socket = Socket, acceptor = Pid}}.

handle_call(_Request, _From, State) ->
	{reply, ok, State}.

handle_cast(_Msg, State) ->
	{noreply, State}.

handle_info({'EXIT', Pid, Reason}, State = #state{acceptor = Pid}) ->
	{stop, {acceptor_died, Reason}, State};
handle_info(_Info, State) ->
	{noreply, State}.

terminate(_Reason, #state{socket = Socket}) ->
	gen_tcp:close(Socket),
	ok.

code_change(_OldVsn, State, _Extra) ->
	{ok, State}.

acceptor_loop(Socket) ->
	case ssl:transport_accept(Socket) of
		{ok, TlsSocket} ->
			case ssl:handshake(TlsSocket) of
				{ok, ClientSocket} ->
					Pid = spawn(
							fun() -> handle_connection(ClientSocket) end),
					ok = ssl:controlling_process(ClientSocket, Pid),
					acceptor_loop(Socket);
				{error, _Reason} ->
					acceptor_loop(Socket)
			end;
		{error, closed} ->
			ok
	end.

handle_connection(Socket) ->
	ssl:setopts(Socket, [{active, true}]),
	handle_connection_loop(Socket).

handle_connection_loop(Socket) ->
	receive
		{ssl, Socket, Data} ->
			handle_client_msg(Socket, Data),
			handle_connection_loop(Socket);
		{ssl_closed, Socket} ->
			ok
	end.

handle_client_msg(Socket, Msg) ->
	ssl:send(Socket, Msg).
